//
// Desencripta los DATA= creados por el AS de PAPI
// Requiere la APR y el OPENSSL
//
// cc -o DecryptDataAS DecryptDataAS.c -lapr-1 -lssl -laprutil-1
//
//

#include <openssl/pem.h>
#include <apr-1.0/apr_strings.h>

#define ASKEYFILE "/etc/apache2/PAPI/UVAS_pubkey.pem"


#define FORMAT_ASN1     1
#define FORMAT_PEM      3

static EVP_PKEY* papi_load_pubkey(char* file, int format) {
        
        BIO* key=NULL;
        EVP_PKEY* pkey=NULL;
        
        if (file == NULL) return NULL;
        
        key=BIO_new(BIO_s_file());

        if (key == NULL) return NULL;
        
        if (BIO_read_filename(key,file) > 0) {          
                if (format == FORMAT_ASN1)
                        pkey=d2i_PUBKEY_bio(key, NULL);
                else if (format == FORMAT_PEM)
                        pkey=PEM_read_bio_PUBKEY(key,NULL,NULL,NULL);
        }
        BIO_free(key);
        return(pkey);
}


char* papi_decrypt_pub_RSA (char* in, char* keyfile)
{
        apr_status_t rv;
        apr_pool_t *pool;

        int blockLength;
        int flen;
        int keysize;
        int n=0;
        char* out=NULL;
        EVP_PKEY *pkey = NULL;
        RSA *rsa = NULL;
        char* rsa_in = NULL;
        int rsa_inlen;
        int rsa_outlen = 0;
        int totalsize=0;

        rv = apr_initialize();
        if (rv != APR_SUCCESS) {
          return apr_pstrdup(pool, "ERRORFILE");
          }
                
        rv = apr_pool_create(&pool, NULL);
                
        pkey = papi_load_pubkey(keyfile, FORMAT_PEM);
        if(!pkey) 
                return apr_pstrdup(pool, "ERRORFILE");

        apr_pool_cleanup_register (pool, pkey, (void *)EVP_PKEY_free, apr_pool_cleanup_null);

        rsa = EVP_PKEY_get1_RSA(pkey);  
        if (!rsa) 
                return apr_pstrdup(pool,"ERRORFILE2");
        apr_pool_cleanup_register (pool, rsa, (void *)RSA_free, apr_pool_cleanup_null);

        rsa_in = apr_pcalloc (pool, apr_base64_decode_len (in)+1);
        flen = apr_base64_decode (rsa_in, in);
        
        keysize = RSA_size(rsa);
        blockLength = keysize;
        
        apr_pool_cleanup_register (pool, out, (void *)free, apr_pool_cleanup_null);
        do {
                rsa_inlen=flen-n;
                rsa_inlen=(rsa_inlen>blockLength)? blockLength:rsa_inlen;
                if (rsa_inlen <= 0) {
                        // APACHE_LOG (APLOG_ERR, "Input empty, rsa_inlen=%d,blockLength=%d,flen=%d",
                        //                           rsa_inlen, blockLength,flen);
                        return apr_pstrdup(pool, "ERRORCRYPT");
                }
                out=(char *)realloc(out,sizeof(char)*(totalsize+keysize+5));
                if (out==NULL) {
                        // APACHE_LOG (APLOG_ERR, "Error reallocating memory");
                        return apr_pstrdup(pool, "ERRORCRYPT");
                }
                rsa_outlen  = RSA_public_decrypt(rsa_inlen, (unsigned char*)&rsa_in[n], (unsigned char*)&out[totalsize], rsa, RSA_PKCS1_PADDING);
                if(rsa_outlen>keysize){
                        // APACHE_LOG (APLOG_ERR, "Length of rsa_out > keysize");
                        return apr_pstrdup (pool,"ERRORCRYPT");
                }
                n+=rsa_inlen;
                totalsize+=rsa_outlen;
        } while (n<flen);
        
        out[totalsize]='\0';
        
        apr_pool_destroy(pool);
                
        return out;
}

main(int argc,char **argv)
{
  if (argc != 2) { printf("Syntax: %s DATA\n",argv[0]); exit(33); }
  
  char *linea = papi_decrypt_pub_RSA (argv[1], ASKEYFILE);
  
  printf ("DATA= %s\n",linea);

  // ERRORFILE
  // ERRORFILE2
  // ERRORCRYPT
  
  return 0;
  
}



